//
// Try this without std::vectors for better performance.
//


#include <iostream>
#include <vector>
#include <complex>
#include <fftw3.h>
#include <boost/algorithm/string.hpp>  // std::max_element()
#include "UDPSimple.hpp"


#define DBG_OUT(x) std::cerr << #x << " = " << x << std::endl



static bool mag_compare(std::complex<float> a, std::complex<float> b)
{
  return (std::abs(a) < std::abs(b));
}

//extern void plot_cf32_buffer(std::string udp_dst_addr, std::string port, void *ptr_cf32, unsigned int n_cf32_values);


class C1DCorrelation {
private:
  unsigned int _signal_length;
  unsigned int _sync_length;
  unsigned int _fft_size;

  float _corr_mag;
  unsigned int _corr_idx;


  std::vector<std::complex<float> > f_signal_buff;  // freq domain signal buffer
  std::vector<std::complex<float> > f_sync_buff;     // freq domain sync buffer

  std::vector<std::complex<float> > f_corr_buff;    // freq domain correlation buffer
  std::vector<std::complex<float> > t_corr_buff;    // time domain correlation buffer

  fftwf_plan ifft_corr_p;

  /*inline*/void find_cf32_max_value_index(std::vector<std::complex<float> > &vec_f32, float &pkm, unsigned int &pki)
  {
    unsigned int length = vec_f32.size();
    for (unsigned int i = 0; i < length; ++i)
      {
	float tabs = abs(vec_f32.at(i));

	if (tabs > pkm)
	  { 
	    pkm = tabs;
	    pki = i;
	  }
      }
  }


public:

  C1DCorrelation() {};

  ~C1DCorrelation() {};


  void init(std::vector<std::complex<float> >& t_sync_buff,
	    unsigned int sync_length,
	    unsigned int signal_length)
  {
    assert(signal_length >= sync_length); // ensure signal length is at least sync length
    assert(sync_length >= t_sync_buff.size()); // ensure sync length matches sync length

    _signal_length = signal_length;
    _sync_length = sync_length;  // some power of 2; 256 // check this is some power of two

    // this will round up sync length to closest power of 2 if needed
    _sync_length = pow(2, ceil(log( _sync_length )/log(2)) );


    _fft_size = pow(2, ceil(log( signal_length )/log(2)) );;

    t_sync_buff.resize( _fft_size );
    f_sync_buff.resize( _fft_size );

    f_signal_buff.resize( _fft_size );
    f_corr_buff.resize( _fft_size );
    t_corr_buff.resize( _fft_size );

    // get fft of sync signal
    fftwf_plan fft_sync_ref_p = fftwf_plan_dft_1d(_fft_size,
						  (fftwf_complex*)&t_sync_buff.front(),
						  (fftwf_complex*)&f_sync_buff.front(),
						  FFTW_FORWARD,
						  FFTW_ESTIMATE);
    fftwf_execute(fft_sync_ref_p);  // should be in f_sync_buff

    fftwf_destroy_plan(fft_sync_ref_p);



    /* fftwf_plan */ ifft_corr_p = fftwf_plan_dft_1d(_fft_size,
					       (fftwf_complex*)&f_corr_buff.front(),
					       (fftwf_complex*)&t_corr_buff.front(),
					       FFTW_BACKWARD,
					       FFTW_ESTIMATE);
    print_setup();
  }

  unsigned int run(std::vector<std::complex<float> >& t_signal_buff)
  {
    
    t_signal_buff.resize(_fft_size);
    fftwf_plan fft_signal_p = fftwf_plan_dft_1d( _fft_size,
					    (fftwf_complex*)&t_signal_buff.front(),
					    (fftwf_complex*)&f_signal_buff.front(),
					    FFTW_FORWARD,
					    FFTW_ESTIMATE);
    // take 512 FFT of signal
    fftwf_execute( fft_signal_p );      // should be in f_signal_buff
    fftwf_destroy_plan( fft_signal_p );

    // plot the signal
    //plot_cf32_buffer("10.10.0.10", "1337",  (void*)t_signal_buff.data(), t_signal_buff.size() );
    
    for (unsigned int i = 0; i < _fft_size; i++)
      f_corr_buff.at(i) = f_signal_buff.at(i) * std::conj(f_sync_buff.at(i));

    fftwf_execute( ifft_corr_p );  // should be in t_corr_buff

    for(unsigned int i = _signal_length; i < t_corr_buff.size()-1; i++) t_corr_buff.at(i) = std::complex<float>(0.0,0.0);

    float corr_mag = -999.0;
    unsigned int corr_idx = 0;
#if 0
    corr_idx = std::distance(t_corr_buff.begin(),
					  std::max_element(t_corr_buff.begin(),
							   t_corr_buff.end(), mag_compare) );
#else
    find_cf32_max_value_index(t_corr_buff, corr_mag, corr_idx);
#endif

    _corr_mag = corr_mag;
    _corr_idx = corr_idx;

    return (corr_idx);

  }



  unsigned int run(fftwf_complex* t_signal_buff_ptr)
  {
    
    fftwf_plan fft_signal_p = fftwf_plan_dft_1d( _fft_size,
					    (fftwf_complex*)t_signal_buff_ptr,
					    (fftwf_complex*)&f_signal_buff.front(),
					    FFTW_FORWARD,
					    FFTW_ESTIMATE);
    // take 512 FFT of signal
    fftwf_execute( fft_signal_p );      // should be in f_signal_buff
    fftwf_destroy_plan( fft_signal_p );

    // plot the signal
    //plot_cf32_buffer("10.10.0.10", "1337",  (void*)t_signal_buff.data(), t_signal_buff.size() );
    
    for (unsigned int i = 0; i < _fft_size; i++)
      f_corr_buff.at(i) = f_signal_buff.at(i) * std::conj(f_sync_buff.at(i));

    fftwf_execute( ifft_corr_p );  // should be in t_corr_buff

    for(unsigned int i = _signal_length; i < t_corr_buff.size()-1; i++) t_corr_buff.at(i) = std::complex<float>(0.0,0.0);

    float corr_mag = -999.0;
    unsigned int corr_idx = 0;
#if 0
    corr_idx = std::distance(t_corr_buff.begin(),
					  std::max_element(t_corr_buff.begin(),
							   t_corr_buff.end(), mag_compare) );
#else
    find_cf32_max_value_index(t_corr_buff, corr_mag, corr_idx);
#endif

    _corr_mag = corr_mag;
    _corr_idx = corr_idx;
    return (corr_idx);

  }

  float corr_mag() { return _corr_mag; }
  float corr_idx() { return _corr_idx; }
  std::vector<std::complex<float> >& corr_magnitudes() { return t_corr_buff; } 


  void  print_setup()
  {
    std::cerr << "Sync signal length:      " << _sync_length << std::endl;
    std::cerr << "Correlation FFT length:  " << _fft_size <<  std::endl;
    std::cerr << "Signal length must be <= " << _fft_size << std::endl;
  }


  unsigned int exec_cf32_sliding_corr( std::vector<std::complex<float> >& t_sync_buff, std::vector<std::complex<float> >& t_signal_buff )
  {
    unsigned int pki = 0;
    float pkm = -999.0;

    assert(t_sync_buff.size() < t_signal_buff.size() );

    unsigned int length = t_signal_buff.size() - t_sync_buff.size() + 1;

    for (unsigned int i = 0; i < length; ++i)
    {
      std::complex<float> cf32 = std::complex<float> (0.0, 0.0);
      for (unsigned int j = 0; j < t_sync_buff.size(); ++j)
	cf32 += t_sync_buff.at(j) * t_signal_buff.at(i+j);
      
      float tabs = abs( cf32 );
      
      if (tabs > pkm)
      {
	pkm = tabs;
	pki = i;
      }
    }
    return pki;
  }

}; // class C1DCorrelation


#if 0

function [x_,xi_, signal_corr_] = fCorr1DSimple(reference_, signal_)


% fft_size = 2 ^ ceil( log2(length(reference_) + length(signal_) ) );
fft_size = 2 ^ ceil( log2( max(length(reference_) , length(signal_) ) ) );

REFERENCE_ = fft(reference_, fft_size);   % this is FD reference signal - done in advance
SIGNAL_  = fft(signal_, fft_size);
signal_corr_ = ifft(SIGNAL_ .* conj(REFERENCE_ ));
[x_,xi_] = max(abs(signal_corr_));


#endif
