#include <iostream>
#include <string.h>
#include <cassert>
//#include <uhd/utils/thread_priority.hpp> // deprecated
#include <uhd/utils/thread.hpp>
#include <uhd/utils/safe_main.hpp>
#include <uhd/usrp/multi_usrp.hpp>
#include <boost/thread/thread.hpp>

#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include "CDeviceStorage.hpp"


#define DBG_OUT(x) std::cerr << #x << " = " << x << std::endl

#define RETURN_SUCCESS 0
#define RETURN_ERROR  -1

// Define number of buffers to store
#define RX_DEV_STORAGE_N_SPB_BUFFS  (512*200)
#define TX_DEV_STORAGE_N_SPB_BUFFS  (10)     // 16 -->  4096 samples / ch
                                             // 64 --> 16384 samples / ch

class CRadio
{
  private:
    uhd::usrp::multi_usrp::sptr _usrp;
    uhd::rx_streamer::sptr _rx_stream;
    uhd::tx_streamer::sptr _tx_stream;

    std::string _args;
    std::string _ref,_sync;
    std::string _subdev;
    std::string _rxant;
    std::string _txant;
    double _rxrate, _rxfreq, _rxgain;
    double _txrate, _txfreq, _txgain;
    double _bw;
    bool _isMakeComplete, _isRxInit, _isTxInit, _is_rx_streaming_on, _is_tx_streaming_on;
    bool _is_rx_pause_on;

    unsigned int _nRxChannels;            // total daughter boards with rx path in this configuration
    unsigned int _nTxChannels;            // total daughter boards with rx path in this configuration
    size_t _spb;
    void _run_rx();
    void _run_tx();
    CDeviceStorage _rx_mdst;
    CDeviceStorage _tx_mdst;
    unsigned long int _rx_total_buffers_read;
    unsigned long int _tx_total_buffers_sent;

  public:
    CRadio();
    int make_device_handle(std::string radio_ip_address);
    CDeviceStorage& rx_mdst() { return _rx_mdst; } 
    CDeviceStorage& tx_mdst() { return _tx_mdst; } 

    unsigned int nRxChannels() { return _nRxChannels; }
    unsigned int nTxChannels() { return _nTxChannels; }
    unsigned int spb()         { return _spb; }

    int run_rx_async( );
    int run_tx_async( );
    bool is_RxStreaming()      { return _is_rx_streaming_on; }
    bool is_TxStreaming()      { return _is_tx_streaming_on; }
    void kill_rx() { _is_rx_streaming_on = false; }
    void kill_tx() { _is_tx_streaming_on = false; }

    bool is_paused() { return _is_rx_pause_on; } 
    void pause_rx() { _is_rx_pause_on = true; }
    void cont_rx()  { _is_rx_pause_on = false; }


    // get handlers for freq gain rate values for a given channel
    double rx_freq(unsigned int ch)   { return _usrp->get_rx_freq( ch ); }
    double rx_rate(unsigned int ch)   { return _usrp->get_rx_rate( ch ); }
    double rx_gain(unsigned int ch)   { return _usrp->get_rx_gain( ch ); }

    double tx_freq(unsigned int ch)   { return _usrp->get_tx_freq( ch ); }
    double tx_rate(unsigned int ch)   { return _usrp->get_tx_rate( ch ); }
    double tx_gain(unsigned int ch)   { return _usrp->get_tx_gain( ch ); }
  


    // set handlers for rx freq gain rate values for a given channel
    void set_sync_pps() { _sync = "pps";  }
  
    double rx_freq(unsigned int ch, double val)
    {
      _usrp->set_rx_freq( val, ch );
      return _usrp->get_rx_freq( ch );
    }

    double rx_rate(unsigned int ch, double val) 
    {
      _usrp->set_rx_rate( val, ch );
      return _usrp->get_rx_rate( ch );
    }
    
    double rx_gain(unsigned int ch, double val)
    {
      _usrp->set_rx_gain( val, ch );
      return _usrp->get_rx_gain( ch );
    }

    double tx_freq(unsigned int ch, double val)
    {
      _usrp->set_tx_freq( val, ch );
      return _usrp->get_tx_freq( ch );
    }

    double tx_rate(unsigned int ch, double val) 
    {
      _usrp->set_tx_rate( val, ch );
      return _usrp->get_tx_rate( ch );
    }
    
    double tx_gain(unsigned int ch, double val)
    {
      _usrp->set_tx_gain( val, ch );
      return _usrp->get_tx_gain( ch );
    }

    std::string get_rx_ant() { return _rxant; }

    std::string mdst_state() {
      std::string temp;
      if (_isMakeComplete == false) {
	temp = "\nNo devices - run 'make' first";
	return (temp);
      }

      temp = "\n[RX]: tot buf read || depth | tail | head : " \
	+ boost::lexical_cast<std::string>(_rx_total_buffers_read ) + " " \
	+ boost::lexical_cast<std::string>(_rx_mdst.depth() ) + " "	\
	+ boost::lexical_cast<std::string>(_rx_mdst.tail() )  + " "	\
	+ boost::lexical_cast<std::string>(_rx_mdst.head() )  + " "	\
	+ "\n        Streaming:         " +  (is_RxStreaming() ? std::string("on") : std::string("off")) \
	+ "\n      samps/buffer  :    " +  boost::lexical_cast<std::string>(_rx_mdst.spb() ) + " " \
	+ "\n      buffers / chan:    " +  boost::lexical_cast<std::string>(_rx_mdst.nbuffptrs() ) + " " \
	+ "\n      tot samps / chan:  " +  boost::lexical_cast<std::string>(_rx_mdst.nsamps_per_ch() ) + " " \
	+ "\n      channels:          " +  boost::lexical_cast<std::string>(_rx_mdst.nChannels() ) + " ";
      for (unsigned int ch = 0; ch < _rx_mdst.nChannels(); ++ch) {
	temp += "\n      channel[" + boost::lexical_cast<std::string>(ch) +"]:" + " " \
	  + boost::lexical_cast<std::string>(_usrp->get_rx_freq( ch ) / 1e6) + " MHz   " \
	  + boost::lexical_cast<std::string>(_usrp->get_rx_rate( ch ) / 1e6) + " Msps   " \
	  + boost::lexical_cast<std::string>(_usrp->get_rx_gain( ch )) + " dB   " \
	  + _usrp->get_rx_antenna(ch);
      }

      temp += "\n[TX]: tot buf sent || depth | tail | head : " \
	+ boost::lexical_cast<std::string>(_tx_total_buffers_sent ) + " " \
	+ boost::lexical_cast<std::string>(_tx_mdst.depth() ) + " "	\
	+ boost::lexical_cast<std::string>(_tx_mdst.tail() )  + " "	\
	+ boost::lexical_cast<std::string>(_tx_mdst.head() )  + " "	\
	+ "\n        Streaming:         " +  (is_TxStreaming() ? std::string("on") : std::string("off")) \
	+ "\n      samps/buffer  :    " +  boost::lexical_cast<std::string>(_tx_mdst.spb() ) + " " \
	+ "\n      buffers / chan:    " +  boost::lexical_cast<std::string>(_tx_mdst.nbuffptrs() ) + " " \
	+ "\n      tot samps / chan:  " +  boost::lexical_cast<std::string>(_tx_mdst.nsamps_per_ch() ) + " " \
	+ "\n      channels:          " +  boost::lexical_cast<std::string>(_tx_mdst.nChannels() ) + " ";
      for (unsigned int ch = 0; ch < _tx_mdst.nChannels(); ++ch) {
	temp += "\n      channel[" + boost::lexical_cast<std::string>(ch) +"]:" + " " \
	  + boost::lexical_cast<std::string>(_usrp->get_tx_freq( ch ) / 1e6) + " MHz   " \
	  + boost::lexical_cast<std::string>(_usrp->get_tx_rate( ch ) / 1e6) + " Msps   " \
	  + boost::lexical_cast<std::string>(_usrp->get_tx_gain( ch )) + " dB   " \
	  + _usrp->get_tx_antenna(ch);
      }

      temp += "\n **bytes / sample:       " +  boost::lexical_cast<std::string>(sizeof(std::complex<float>)) + " ";

      return (temp);
    }

};




CRadio::CRadio()
{
  _isRxInit = _isTxInit = _is_rx_streaming_on = _is_tx_streaming_on =_isMakeComplete = false;
  _tx_stream = NULL;
  _rx_stream = NULL;
  _sync = "now";
  //_sync = "pps";
  //_rxant = "RX2";
  _rxant = "TX/RX";
  _txant = "TX/RX";
  _bw = 25e6;
  _spb = 256;
  _rx_total_buffers_read = _tx_total_buffers_sent = 0;
}

int CRadio::make_device_handle(std::string radio_ip_address)
{
  if (_isMakeComplete == true) return RETURN_ERROR;

  _args = radio_ip_address;

  // =====================================================================================
  // Create multi-usrp device handler
  // =====================================================================================
  std::cerr << boost::format("Creating the usrp device: %s") % _args << std::endl;

  _usrp = uhd::usrp::multi_usrp::make(_args);
  std::cerr << boost::format("Using Device: %s") % _usrp->get_pp_string() << std::endl;

  unsigned int num_mboards = _usrp->get_num_mboards();
  _nRxChannels = _usrp->get_rx_num_channels();
  _nTxChannels = _usrp->get_tx_num_channels();
  
  std::cerr << "Number mboards        = " << num_mboards << std::endl;
  std::cerr << "Number of rx channels = " << _nRxChannels << std::endl;
  std::cerr << "Number of tx channels = " << _nTxChannels << std::endl;

  // =====================================================================================
  // Device synchronization
  // =====================================================================================
  std::cerr << boost::format("Setting device timestamp to 0...") << std::endl;
  if (_sync == "now"){
    _ref = "internal";
    //This is not a true time lock, the devices will be off by a few RTT.
    //Rather, this is just to allow for demonstration of the code below.
    _usrp->set_clock_source(_ref);
    _usrp->set_time_now(uhd::time_spec_t(0.0) );
  }
  else if (_sync == "pps"){
    _ref = "external";
    _usrp->set_time_source(_ref);
    _usrp->set_clock_source(_ref);
    _usrp->set_time_unknown_pps(uhd::time_spec_t(0.0));
    boost::this_thread::sleep(boost::posix_time::seconds(1)); //wait for pps sync pulse
  }
  else if (_sync == "mimo"){
    UHD_ASSERT_THROW(_usrp->get_num_mboards() == 2);

    //make mboard 1 a slave over the MIMO Cable
    _usrp->set_clock_source("mimo", 1);
    _usrp->set_time_source("mimo", 1);

    //set time on the master (mboard 0)
    _usrp->set_time_now(uhd::time_spec_t(0.0), 0);

    //sleep a bit while the slave locks its time to the master
    boost::this_thread::sleep(boost::posix_time::milliseconds(100));
  }


  // =====================================================================================
  // Device parameters: {rx,tx} freq, gain, rate, antenna, bw
  // =====================================================================================
  for (unsigned int chidx = 0; chidx < _nRxChannels; ++chidx) {
    _usrp->set_rx_antenna(   _rxant, chidx );
    _usrp->set_rx_bandwidth( _bw, chidx );
  }

  for (unsigned int chidx = 0; chidx < _nTxChannels; ++chidx) {
    _usrp->set_tx_antenna(   _txant, chidx );
    _usrp->set_tx_bandwidth( _bw, chidx );
  }


  // =====================================================================================
  // Storage for transmit & received samples.
  // =====================================================================================
  _tx_mdst.init( _nTxChannels , _spb, TX_DEV_STORAGE_N_SPB_BUFFS);
  _tx_mdst.print_info("TX");

  _rx_mdst.init( _nRxChannels , _spb, RX_DEV_STORAGE_N_SPB_BUFFS);
  _rx_mdst.print_info("RX");

  _isMakeComplete = true;

  return RETURN_SUCCESS;

}


int CRadio::run_rx_async()
{
  if ( _is_rx_streaming_on ) return RETURN_ERROR;

  boost::thread t(boost::bind( &CRadio::_run_rx, this ) );
  _is_rx_streaming_on = true;

}

void CRadio::_run_rx()
{
  _usrp->set_time_now(uhd::time_spec_t(0.0) );

  //create a receive streamer - populate channel list
  //linearly map channels (index0 = channel0, index1 = channel1, ...)
  uhd::stream_args_t stream_args("fc32"); //complex floats
  for (unsigned int i = 0; i < _nRxChannels; i++)
    stream_args.channels.push_back(i);
  assert( stream_args.channels.size() == _nRxChannels);

  _rx_stream = _usrp->get_rx_stream(stream_args);

  //setup streaming
  double delay_sec = 0.5;
  uhd::stream_cmd_t stream_cmd(uhd::stream_cmd_t::STREAM_MODE_START_CONTINUOUS);
  stream_cmd.stream_now = false;
  stream_cmd.time_spec = uhd::time_spec_t(delay_sec);
  _rx_stream->issue_stream_cmd(stream_cmd); //tells all channels to stream samples to host. This is the GO button.


  //the first call to recv() will block this many seconds before receiving
  double timeout = delay_sec + 0.1; //timeout (delay before receive + padding)
  boost::system_time next_refresh = boost::get_system_time();
  uhd::rx_metadata_t _md;
  size_t num_rx_samps;
  //float pkt_time_intv_us = 1/_usrp->get_rx_rate( 0 ) * _spb * 1e6;
  //DBG_OUT(pkt_time_intv_us) ;

  DBG_OUT(_is_rx_streaming_on) ;

  while( _is_rx_streaming_on ) {

    // if (boost::get_system_time() < next_refresh) continue;
    // next_refresh = boost::get_system_time() + boost::posix_time::microseconds(long( pkt_time_intv_us  ));

    //receive multi channel buffers
    num_rx_samps = _rx_stream->recv( _rx_mdst.buffer_ptr_idx(_rx_mdst.head() ), _spb, _md, timeout);

    if (num_rx_samps != _spb) {
      std::cerr << "!!";
      continue;
    }
    _rx_total_buffers_read++;

    //handle the error code
    if (_md.error_code == uhd::rx_metadata_t::ERROR_CODE_TIMEOUT) break;
    else if (_md.error_code == uhd::rx_metadata_t::ERROR_CODE_OVERFLOW) continue;
    else if (_md.error_code != uhd::rx_metadata_t::ERROR_CODE_NONE){
      throw std::runtime_error(str(boost::format("Recv'd samples %i\nReceiver error %s") % num_rx_samps % _md.strerror()));
    }
    
    //use a small timeout for subsequent packets
    timeout = 0.1;

    if (_is_rx_pause_on) continue;

    _rx_mdst.time( _md.time_spec.get_full_secs(), _md.time_spec.get_frac_secs());


    // Select behavior when storage buffers are full
    // Option 1 -  drop from tail end 
    if (_rx_mdst.head( +1 ) == _rx_mdst.tail(  ) ) _rx_mdst.tail_inc( );

    // Option 2 - hold head from incrementing  
    //if (_rx_mdst.head( +1 ) == _rx_mdst.tail(  ) ) continue;


    _rx_mdst.head_inc( );

  } // while()
  DBG_OUT(_is_rx_streaming_on) ;

  stream_cmd.stream_mode = uhd::stream_cmd_t::STREAM_MODE_STOP_CONTINUOUS;
  stream_cmd.stream_now = true;
  _rx_stream->issue_stream_cmd(stream_cmd);

} // end of run_rx








int CRadio::run_tx_async()
{
  if ( _is_tx_streaming_on ) return RETURN_ERROR;

  boost::thread t(boost::bind( &CRadio::_run_tx, this ));
  _is_tx_streaming_on = true;

  return RETURN_SUCCESS;

}


void CRadio::_run_tx()
{

  std::vector<std::complex<float> > zero_buf(_spb);
  std::vector<std::complex<float> *> zero_buff_ptrs;
  for (size_t i = 0; i < _nTxChannels; i++) zero_buff_ptrs.push_back(&zero_buf.front() );

  //create a receive streamer - populate channel list
  //linearly map channels (index0 = channel0, index1 = channel1, ...)
  uhd::stream_args_t stream_args("fc32"); //complex floats
  for (unsigned int i = 0; i < _nTxChannels; i++)
    stream_args.channels.push_back(i);
  assert( stream_args.channels.size() == _nTxChannels);


  // only get tx stream is this is first time run_tx
  if (_tx_stream == NULL)
    _tx_stream = _usrp->get_tx_stream(stream_args);

  //setup streaming
  double delay_sec = 1.5;

  //the first call to recv() will block this many seconds before receiving
  double timeout = delay_sec + 0.1; //timeout (delay before receive + padding)
  //boost::system_time next_console_refresh = boost::get_system_time();
  boost::system_time next_refresh = boost::get_system_time();

  uhd::tx_metadata_t md;   //setup the metadata flags
  md.start_of_burst = true;
  md.end_of_burst   = false;
  md.has_time_spec  = true;
  md.time_spec = uhd::time_spec_t(0.1);
  
  //float pkt_time_intv_us = 1/_usrp->get_tx_rate( 0 ) * _spb * 1e6;
  //pkt_time_intv_us = 0.0;

  //DBG_OUT(pkt_time_intv_us);
  DBG_OUT(_is_tx_streaming_on);
  while( _is_tx_streaming_on ) {
    
    //if (boost::get_system_time() < next_refresh) continue;
    //next_refresh = boost::get_system_time() + boost::posix_time::microseconds(long( pkt_time_intv_us  ));

    if (_tx_mdst.head() == 0) {
      _tx_stream->send(zero_buff_ptrs, _spb, md);
      md.start_of_burst = false;
      md.has_time_spec = false;
      continue;
    }

    // send multi channel buffers to device
    for (unsigned int idx = 0; idx < _tx_mdst.head(); idx++)
    {
      _tx_stream->send(_tx_mdst.buffer_ptr_idx( idx ), _spb, md);
      md.start_of_burst = false;
      md.has_time_spec = false;
    }

    _tx_total_buffers_sent++;
    
    // Comment line below to keep sending sequence
    _tx_mdst.head_set(0);

  } // while()

  // ???? send a mini EOB packet
  md.end_of_burst = true;
  _tx_stream->send("", 0, md);

}
